#include <vector>

#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/math/special_functions/sign.hpp>

#include "Board.hpp"
#include "Exceptions.hpp"

Board::Board()
{
    // none
}

Board Board::from_fen(std::string fen) {
    std::vector<std::string> fen_parts;
    boost::split(fen_parts, fen, boost::is_space());

    std::string board = fen_parts.at(0);
    std::string turn_color = fen_parts.at(1);
    std::string rooking_posibility = fen_parts.at(2);
    std::string enpassant_posibility = fen_parts.at(3);
    std::string unused_moves_count = fen_parts.at(4);
    std::string moves_count = fen_parts.at(5);

    Board result;

    result.turn_color = Color::from_fen(turn_color);
    result.rooking_posibility = RookingPosibility::from_fen(rooking_posibility);
    result.enpassant_posibility = EnpassantPosibility::from_fen(enpassant_posibility);
    result.unused_moves_count = boost::lexical_cast<UNumber>(unused_moves_count);
    result.moves_count = boost::lexical_cast<UNumber>(moves_count);

    UNumber file = 1;
    UNumber rank = 8;
    for (char ch: board) {
        if (ch == '/') {
            --rank;
            file = 1;
            continue;
        }

        if ('1' <= ch && ch <= '9') {
            file += (ch - '0');
            continue;
        }

        result.add_piece( Piece::from_fen(boost::lexical_cast<std::string>(ch)), Coord(file, rank) );

        file++;
    }

    return result;
}

Board Board::initial() {
    return Board::from_fen("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
}

Board Board::empty() {
    return Board::from_fen("8/8/8/8/8/8/8/8 w - - 0 1");
}

void Board::add_piece( const Piece& piece, const Coord& coord ) {
    if ((*this).has_piece_at(coord))
        throw CustomException("Trying to add piece in nonempty square");

    (*this).squares.insert(std::pair<Coord, Piece>(coord, piece));
}

void Board::remove_piece( const Coord& coord ) {
    if (!(*this).has_piece_at(coord))
        throw CustomException("Trying to remove piece from empty square");

    (*this).squares.erase(coord);
}

void Board::move_piece( const Coord& from, const Coord& to ) {
    (*this).add_piece(squares.at(from), to);
    (*this).remove_piece(from);
}

bool Board::has_piece_at( const Coord& coord ) const {
    return (*this).squares.find(coord) != (*this).squares.end();
}

Piece Board::get_piece_at( const Coord& coord ) const {
	if (!has_piece_at(coord)) throw CustomException("Trying to get piece from empty square");

	Piece p = (*this).squares.find(coord)->second;
    return p;
}

bool Board::line_is_empty( const Coord& from, const Coord& to ) const {
    const SNumber direction_file = boost::math::sign((SNumber)to.file - (SNumber)from.file);
    const SNumber direction_rank = boost::math::sign((SNumber)to.rank - (SNumber)from.rank);

    for (SNumber delta = 1; delta <= 8; ++delta) {
        if (!Coord::is_valid(from.file+delta*direction_file, from.rank+delta*direction_rank)) break;

        auto temporary = Coord(from.file+delta*direction_file, from.rank+delta*direction_rank);
        if (temporary == to) break;

        if ((*this).has_piece_at(temporary))
            return false;
    }

    return true;
}

bool Board::is_empty() const {
    return (*this).squares.empty();
}

UNumber Board::pieces_count() const {
    return (*this).squares.size();
}

std::string Board::to_fen() const {
    std::string result;

    for (UNumber rank = 8; rank >= 1; --rank) {
        UNumber empty_count = 0;
        for (UNumber file = 1; file <= 8; ++file) {
            if ( !(*this).has_piece_at(Coord(file, rank)) )
                ++empty_count;
            else {
                if (empty_count != 0) {
                    result += '0' + empty_count;
                    empty_count = 0;
                }
                result += (*this).squares.at(Coord(file, rank)).to_fen();
            }
        }
        if (empty_count != 0) {
            result += '0' + empty_count;
            empty_count = 0;
        }

        if (rank != 1) result += '/';
    }

    result += " " + (*this).turn_color.to_fen();
    result += " " + (*this).rooking_posibility.to_fen();
    result += " " + (*this).enpassant_posibility.to_fen();
    result += " " + boost::lexical_cast<std::string>((*this).unused_moves_count);
    result += " " + boost::lexical_cast<std::string>((*this).moves_count);

    return result;
}

bool Board::operator==(const Board& rhs) const {
    return (*this).squares == rhs.squares;
}

bool Board::operator!=(const Board& rhs) const {
    return (*this).squares != rhs.squares;
}
