#include <boost/random.hpp>
#include <limits>

#include "Hash.hpp"
#include "Exceptions.hpp"
#include "Search.hpp"

std::map<std::pair<Coord, Piece>, UNumber64> zorbist_square_table;
std::map<Color, UNumber64> zorbist_turn_table;

void init_zorbist_tables() {
	static bool initiated = false;
	if (initiated) {
		return;
	}
	initiated = true;

	boost::mt19937 random_generator;
	random_generator.seed(time(0));
	boost::uniform_int<UNumber64> range(0, std::numeric_limits<UNumber64>::max());
	boost::variate_generator<boost::mt19937, boost::uniform_int<UNumber64>> random64(random_generator, range);

	for (auto coord : Coord::range) {
		for (auto piece : Piece::range) {
			auto square = std::pair<Coord, Piece>(coord, piece);
			zorbist_square_table.insert(std::pair<std::pair<Coord, Piece>, UNumber64>(square, random64()));
		}
	}

	zorbist_turn_table.insert(std::pair<Color, UNumber64>(WHITE, random64()));
	zorbist_turn_table.insert(std::pair<Color, UNumber64>(BLACK, random64()));
}

UNumber64 Hash::get64(const Board& board) {
	init_zorbist_tables();

    UNumber64 result = 0;

    for (auto square: board.squares) {
		result ^= zorbist_square_table.at(square);
    }

	result ^= zorbist_turn_table.at(board.turn_color);

    return result;
}

void Hash::Table::add(const Board& board, Hash::Record record) {
	if ((*this).has(board)) throw CustomException("Rewriting record at hash");

	auto key = get64(board);

	(*this).records.insert(std::pair<UNumber64, Record>(key, record));
}

void Hash::Table::renew(const Board& board, Hash::Record new_rec) {
    if (Search::Manager::stoped) return;

	if (!(*this).has(board)) (*this).add(board, new_rec);
	else {
		auto old_rec = (*this).get(board);
		if (new_rec.depth >= old_rec.depth || (new_rec.nature != old_rec.nature && new_rec.nature == Record::SCORE_REAL)) {
			(*this).remove(board);
			(*this).add(board, new_rec);
		}
		if (new_rec.depth == old_rec.depth && new_rec.nature == old_rec.nature) {
			if (new_rec.nature == Record::SCORE_MAX && new_rec.score < old_rec.score) {
				(*this).remove(board);
				(*this).add(board, new_rec);
			}
			if (new_rec.nature == Record::SCORE_MIN && new_rec.score > old_rec.score) {
				(*this).remove(board);
				(*this).add(board, new_rec);
			}
		}
	}
}

bool Hash::Table::has(const Board& board) {
	auto key = get64(board);

	return (*this).records.find(key) != (*this).records.end();
}

void Hash::Table::remove(const Board& board) {
	auto key = get64(board);

	(*this).records.erase(key);
}

Hash::Record Hash::Table::get(const Board& board) {
	if (!(*this).has(board)) throw CustomException("Getting empty record from hash");

	auto key = get64(board);

	return (*this).records.at(key);
}

UNumber Hash::Table::size() {
	return (*this).records.size();
}

Hash::Table Hash::Table::empty() {
	Table result;
	return result;
}
